@@ -95,6 +95,7 @@ PHP_INI_BEGIN()
 	PHP_INI_ENTRY("amqp.cert",				DEFAULT_CERT,				PHP_INI_ALL, NULL)
 	PHP_INI_ENTRY("amqp.key",				DEFAULT_KEY,				PHP_INI_ALL, NULL)
 	PHP_INI_ENTRY("amqp.verify",			DEFAULT_VERIFY,				PHP_INI_ALL, NULL)
+	PHP_INI_ENTRY("amqp.sasl_method",		DEFAULT_SASL_METHOD,		PHP_INI_ALL, NULL)
 PHP_INI_END()
 
 ZEND_DECLARE_MODULE_GLOBALS(amqp);
@@ -172,6 +173,8 @@ static PHP_MINIT_FUNCTION(amqp) /* {{{ */
 	REGISTER_LONG_CONSTANT("AMQP_OS_SOCKET_TIMEOUT_ERRNO",	AMQP_OS_SOCKET_TIMEOUT_ERRNO,	CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("PHP_AMQP_MAX_CHANNELS",			PHP_AMQP_MAX_CHANNELS,			CONST_CS | CONST_PERSISTENT);
 
+	REGISTER_LONG_CONSTANT("AMQP_SASL_METHOD_PLAIN",		AMQP_SASL_METHOD_PLAIN,		CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("AMQP_SASL_METHOD_EXTERNAL",		AMQP_SASL_METHOD_EXTERNAL,	CONST_CS | CONST_PERSISTENT);
 	return SUCCESS;
 } /* }}} */
 
@@ -184,14 +187,14 @@ static PHP_MSHUTDOWN_FUNCTION(amqp) /* {{{ */
 
 static PHP_RSHUTDOWN_FUNCTION(amqp) /* {{{ */
 {
-    if (NULL != PHP_AMQP_G(error_message)) {
+	if (NULL != PHP_AMQP_G(error_message)) {
 		efree(PHP_AMQP_G(error_message));
 		PHP_AMQP_G(error_message) = NULL;
-    }
+	}
 
-    PHP_AMQP_G(error_code) = 0;
+	PHP_AMQP_G(error_code) = 0;
 
-    return SUCCESS;
+	return SUCCESS;
 } /* }}} */
 
 static PHP_MINFO_FUNCTION(amqp) /* {{{ */
