@@ -21,8 +21,6 @@
   +----------------------------------------------------------------------+
 */
 
-/* $Id: amqp_connection.c 327551 2012-09-09 03:49:34Z pdezwart $ */
-
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -40,9 +38,11 @@
 # include <signal.h>
 # include <stdint.h>
 #endif
+
 #include <amqp.h>
-#include <amqp_framing.h>
 #include <amqp_tcp_socket.h>
+#include <amqp_framing.h>
+#include <amqp_ssl_socket.h>
 
 #ifdef PHP_WIN32
 # include "win32/unistd.h"
@@ -50,9 +50,11 @@
 # include <unistd.h>
 #endif
 
-#include "php_amqp.h"
-#include "amqp_channel.h"
+//#include "amqp_basic_properties.h"
+#include "amqp_methods_handling.h"
 #include "amqp_connection_resource.h"
+#include "amqp_channel.h"
+#include "php_amqp.h"
 
 #ifndef E_DEPRECATED
 #define E_DEPRECATED E_WARNING
@@ -62,10 +64,11 @@ int le_amqp_connection_resource;
 int le_amqp_connection_resource_persistent;
 
 static void connection_resource_destructor(amqp_connection_resource *resource, int persistent TSRMLS_DC);
+static void php_amqp_close_connection_from_server(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource TSRMLS_DC);
+static void php_amqp_close_channel_from_server(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource, amqp_channel_t channel_id TSRMLS_DC);
 
 
 /* Figure out what's going on connection and handle protocol exceptions, if any */
-
 int php_amqp_connection_resource_error(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource, amqp_channel_t channel_id TSRMLS_DC)
 {
 	assert (resource != NULL);
@@ -79,71 +82,18 @@ int php_amqp_connection_resource_error(amqp_rpc_reply_t reply, char **message, a
 			return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
 
 		case AMQP_RESPONSE_LIBRARY_EXCEPTION:
+
 			spprintf(message, 0, "Library error: %s", amqp_error_string2(reply.library_error));
 			return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
 
 		case AMQP_RESPONSE_SERVER_EXCEPTION:
 			switch (reply.reply.id) {
 				case AMQP_CONNECTION_CLOSE_METHOD: {
-					amqp_connection_close_t *m = (amqp_connection_close_t *)reply.reply.decoded;
-
-					spprintf(message, 0, "Server connection error: %d, message: %.*s",
-						m->reply_code,
-						(PHP5to7_param_str_len_type_t) m->reply_text.len,
-						(char *) m->reply_text.bytes
-					);
-
-					/*
-					 *    - If r.reply.id == AMQP_CONNECTION_CLOSE_METHOD a connection exception
-					 *      occurred, cast r.reply.decoded to amqp_connection_close_t* to see
-					 *      details of the exception. The client amqp_send_method() a
-					 *      amqp_connection_close_ok_t and disconnect from the broker.
-					 */
-
-					amqp_connection_close_ok_t *decoded = (amqp_connection_close_ok_t *) NULL;
-
-					amqp_send_method(
-						resource->connection_state,
-						0, /* NOTE: 0-channel is reserved for things like this */
-						AMQP_CONNECTION_CLOSE_OK_METHOD,
-						&decoded
-					);
-
-					/* Prevent finishing AMQP connection in connection resource destructor */
-					resource->is_connected = '\0';
-
+					php_amqp_close_connection_from_server(reply, message, resource TSRMLS_CC);
 					return PHP_AMQP_RESOURCE_RESPONSE_ERROR_CONNECTION_CLOSED;
 				}
 				case AMQP_CHANNEL_CLOSE_METHOD: {
-					assert(channel_id > 0 && channel_id <= resource->max_slots);
-
-					amqp_channel_close_t *m = (amqp_channel_close_t *) reply.reply.decoded;
-
-					spprintf(message, 0, "Server channel error: %d, message: %.*s",
-						m->reply_code,
-						(PHP5to7_param_str_len_type_t) m->reply_text.len,
-						(char *)m->reply_text.bytes
-					);
-
-					/*
-					 *    - If r.reply.id == AMQP_CHANNEL_CLOSE_METHOD a channel exception
-					 *      occurred, cast r.reply.decoded to amqp_channel_close_t* to see details
-					 *      of the exception. The client should amqp_send_method() a
-					 *      amqp_channel_close_ok_t. The channel must be re-opened before it
-					 *      can be used again. Any resources associated with the channel
-					 *      (auto-delete exchanges, auto-delete queues, consumers) are invalid
-					 *      and must be recreated before attempting to use them again.
-					 */
-
-					amqp_channel_close_ok_t *decoded = (amqp_channel_close_ok_t *) NULL;
-
-					amqp_send_method(
-						resource->connection_state,
-						channel_id,
-						AMQP_CHANNEL_CLOSE_OK_METHOD,
-						&decoded
-					);
-
+					php_amqp_close_channel_from_server(reply, message, resource, channel_id TSRMLS_CC);
 					return PHP_AMQP_RESOURCE_RESPONSE_ERROR_CHANNEL_CLOSED;
 				}
 			}
@@ -156,6 +106,165 @@ int php_amqp_connection_resource_error(amqp_rpc_reply_t reply, char **message, a
 	/* Should not never get here*/
 }
 
+static void php_amqp_close_connection_from_server(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource TSRMLS_DC) {
+	amqp_connection_close_t *m = (amqp_connection_close_t *)reply.reply.decoded;
+	int result;
+
+	if (!reply.reply.id) {
+		PHP_AMQP_G(error_code) = -1;
+		spprintf(message, 0, "Server connection error: %d, message: %s",
+				 PHP_AMQP_G(error_code),
+				 "unexpected response"
+		);
+	} else {
+		PHP_AMQP_G(error_code) = m->reply_code;
+		spprintf(message, 0, "Server connection error: %d, message: %.*s",
+				 m->reply_code,
+				 (PHP5to7_param_str_len_type_t) m->reply_text.len,
+				 (char *) m->reply_text.bytes
+		);
+	}
+		
+	/*
+	 *    - If r.reply.id == AMQP_CONNECTION_CLOSE_METHOD a connection exception
+	 *      occurred, cast r.reply.decoded to amqp_connection_close_t* to see
+	 *      details of the exception. The client amqp_send_method() a
+	 *      amqp_connection_close_ok_t and disconnect from the broker.
+	 */
+
+	amqp_connection_close_ok_t *decoded = (amqp_connection_close_ok_t *) NULL;
+
+	result = amqp_send_method(
+		resource->connection_state,
+		0, /* NOTE: 0-channel is reserved for things like this */
+		AMQP_CONNECTION_CLOSE_OK_METHOD,
+		&decoded
+	);
+
+	if (result != AMQP_STATUS_OK) {
+		zend_throw_exception(amqp_channel_exception_class_entry, "An error occurred while closing the connection.", 0 TSRMLS_CC);
+	}
+
+	/* Prevent finishing AMQP connection in connection resource destructor */
+	resource->is_connected = '\0';
+}
+
+static void php_amqp_close_channel_from_server(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource, amqp_channel_t channel_id TSRMLS_DC) {
+	assert(channel_id > 0 && channel_id <= resource->max_slots);
+
+	amqp_channel_close_t *m = (amqp_channel_close_t *) reply.reply.decoded;
+
+	if (!reply.reply.id) {
+		PHP_AMQP_G(error_code) = -1;
+		spprintf(message, 0, "Server channel error: %d, message: %s",
+				 PHP_AMQP_G(error_code),
+				 "unexpected response"
+		);
+	} else {
+		PHP_AMQP_G(error_code) = m->reply_code;
+		spprintf(message, 0, "Server channel error: %d, message: %.*s",
+			m->reply_code,
+			(PHP5to7_param_str_len_type_t) m->reply_text.len,
+			(char *)m->reply_text.bytes
+		);
+	}
+
+	/*
+	 *    - If r.reply.id == AMQP_CHANNEL_CLOSE_METHOD a channel exception
+	 *      occurred, cast r.reply.decoded to amqp_channel_close_t* to see details
+	 *      of the exception. The client should amqp_send_method() a
+	 *      amqp_channel_close_ok_t. The channel must be re-opened before it
+	 *      can be used again. Any resources associated with the channel
+	 *      (auto-delete exchanges, auto-delete queues, consumers) are invalid
+	 *      and must be recreated before attempting to use them again.
+	 */
+
+	if (resource) {
+		int result;
+		amqp_channel_close_ok_t *decoded = (amqp_channel_close_ok_t *) NULL;
+
+		result = amqp_send_method(
+				resource->connection_state,
+				channel_id,
+				AMQP_CHANNEL_CLOSE_OK_METHOD,
+				&decoded
+		);
+		if (result != AMQP_STATUS_OK) {
+			zend_throw_exception(amqp_channel_exception_class_entry, "An error occurred while closing channel.", 0 TSRMLS_CC);
+		}
+	}
+}
+
+
+int php_amqp_connection_resource_error_advanced(amqp_rpc_reply_t reply, char **message, amqp_connection_resource *resource, amqp_channel_t channel_id, amqp_channel_object *channel TSRMLS_DC)
+{
+	assert(resource != NULL);
+
+	amqp_frame_t frame;
+
+	assert(AMQP_RESPONSE_LIBRARY_EXCEPTION == reply.reply_type);
+	assert(AMQP_STATUS_UNEXPECTED_STATE == reply.library_error);
+
+	if (channel_id < 0 || AMQP_STATUS_OK != amqp_simple_wait_frame(resource->connection_state, &frame)) {
+		if (*message != NULL) {
+			efree(*message);
+		}
+
+		spprintf(message, 0, "Library error: %s", amqp_error_string2(reply.library_error));
+		return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
+	}
+
+	if (channel_id != frame.channel) {
+		spprintf(message, 0, "Library error: channel mismatch");
+		return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
+	}
+
+	if (AMQP_FRAME_METHOD == frame.frame_type) {
+		switch (frame.payload.method.id) {
+			case AMQP_CONNECTION_CLOSE_METHOD: {
+				php_amqp_close_connection_from_server(reply, message, resource TSRMLS_CC);
+				return PHP_AMQP_RESOURCE_RESPONSE_ERROR_CONNECTION_CLOSED;
+			}
+			case AMQP_CHANNEL_CLOSE_METHOD: {
+				php_amqp_close_channel_from_server(reply, message, resource, channel_id TSRMLS_CC);
+				return PHP_AMQP_RESOURCE_RESPONSE_ERROR_CHANNEL_CLOSED;
+			}
+
+			case AMQP_BASIC_ACK_METHOD:
+				/* if we've turned publisher confirms on, and we've published a message
+				 * here is a message being confirmed
+				 */
+
+				return php_amqp_handle_basic_ack(message, resource, channel_id, channel, &frame.payload.method TSRMLS_CC);
+			case AMQP_BASIC_NACK_METHOD:
+				/* if we've turned publisher confirms on, and we've published a message
+				 * here is a message being confirmed
+				 */
+
+				return php_amqp_handle_basic_nack(message, resource, channel_id, channel, &frame.payload.method TSRMLS_CC);
+			case AMQP_BASIC_RETURN_METHOD:
+				/* if a published message couldn't be routed and the mandatory flag was set
+				 * this is what would be returned. The message then needs to be read.
+				 */
+
+				return php_amqp_handle_basic_return(message, resource, channel_id, channel, &frame.payload.method TSRMLS_CC);
+			default:
+				if (*message != NULL) {
+					efree(*message);
+				}
+
+				spprintf(message, 0, "Library error: An unexpected method was received 0x%08X\n", frame.payload.method.id);
+				return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
+		}
+	}
+
+	if (*message != NULL) {
+		efree(*message);
+	}
+
+	spprintf(message, 0, "Library error: %s", amqp_error_string2(reply.library_error));
+	return PHP_AMQP_RESOURCE_RESPONSE_ERROR;
+}
 
 /* Socket-related functions */
 int php_amqp_set_resource_read_timeout(amqp_connection_resource *resource, double timeout TSRMLS_DC)
@@ -296,7 +405,44 @@ amqp_connection_resource *connection_resource_constructor(amqp_connection_params
 	resource->connection_state = amqp_new_connection();
 
 	/* Create socket object */
-	resource->socket = amqp_tcp_socket_new(resource->connection_state);
+	if (params->cacert) {
+		resource->socket = amqp_ssl_socket_new(resource->connection_state);
+
+		if (!resource->socket) {
+			zend_throw_exception(amqp_connection_exception_class_entry, "Socket error: could not create SSL socket.", 0 TSRMLS_CC);
+
+			return NULL;
+		}
+	} else {
+		resource->socket = amqp_tcp_socket_new(resource->connection_state);
+
+		if (!resource->socket) {
+			zend_throw_exception(amqp_connection_exception_class_entry, "Socket error: could not create socket.", 0 TSRMLS_CC);
+
+			return NULL;
+		}
+	}
+
+	if (params->cacert && amqp_ssl_socket_set_cacert(resource->socket, params->cacert)) {
+		zend_throw_exception(amqp_connection_exception_class_entry, "Socket error: could not set CA certificate.", 0 TSRMLS_CC);
+
+		return NULL;
+	}
+
+	if (params->cacert) {
+#if AMQP_VERSION_MAJOR * 100 + AMQP_VERSION_MINOR * 10 + AMQP_VERSION_PATCH >= 80
+		amqp_ssl_socket_set_verify_peer(resource->socket, params->verify);
+		amqp_ssl_socket_set_verify_hostname(resource->socket, params->verify);
+#else
+		amqp_ssl_socket_set_verify(resource->socket, params->verify);
+#endif
+	}
+
+	if (params->cert && params->key && amqp_ssl_socket_set_key(resource->socket, params->cert, params->key)) {
+		zend_throw_exception(amqp_connection_exception_class_entry, "Socket error: could not setting client cert.", 0 TSRMLS_CC);
+
+		return NULL;
+	}
 
 	if (params->connect_timeout > 0) {
 		tv.tv_sec = (long int) params->connect_timeout;
@@ -368,20 +514,20 @@ amqp_connection_resource *connection_resource_constructor(amqp_connection_params
 		params->frame_max,
 		params->heartbeat,
 		&custom_properties_table,
-		AMQP_SASL_METHOD_PLAIN,
+		params->sasl_method,
 		params->login,
 		params->password
 	);
 
 	efree(std_datetime);
 
-	if (res.reply_type != AMQP_RESPONSE_NORMAL) {
-		char *message, *long_message;
+	if (AMQP_RESPONSE_NORMAL != res.reply_type) {
+		char *message = NULL, *long_message = NULL;
 
 		php_amqp_connection_resource_error(res, &message, resource, 0 TSRMLS_CC);
 
 		spprintf(&long_message, 0, "%s - Potential login failure.", message);
-		zend_throw_exception(amqp_connection_exception_class_entry, long_message, 0 TSRMLS_CC);
+		zend_throw_exception(amqp_connection_exception_class_entry, long_message, PHP_AMQP_G(error_code) TSRMLS_CC);
 
 		efree(message);
 		efree(long_message);
@@ -401,7 +547,7 @@ amqp_connection_resource *connection_resource_constructor(amqp_connection_params
 	}
 
 	/* Allocate space for the channel slots in the ring buffer */
-    resource->max_slots = (amqp_channel_t) amqp_get_channel_max(resource->connection_state);
+	resource->max_slots = (amqp_channel_t) amqp_get_channel_max(resource->connection_state);
 	assert(resource->max_slots > 0);
 
 	resource->slots = (amqp_channel_resource **)pecalloc(resource->max_slots + 1, sizeof(amqp_channel_object*), persistent);
@@ -484,7 +630,7 @@ void php_amqp_prepare_for_disconnect(amqp_connection_resource *resource TSRMLS_D
 
 		for (slot = 0; slot < resource->max_slots; slot++) {
 			if (resource->slots[slot] != 0) {
-				php_amqp_close_channel(resource->slots[slot] TSRMLS_CC);
+				php_amqp_close_channel(resource->slots[slot], 0 TSRMLS_CC);
 			}
 		}
 	}
